from django.apps import AppConfig

from .models import BusinessRuleSet


class DynamicBusinessRulesConfig(AppConfig):
    name = 'dynamic_business_rules'

    def ready(self):
        for ruleset in BusinessRuleSet.objects.all:
            ruleset.signal_object.connect(ruleset.signal_callback)



# Generated by Django 2.0.7 on 2018-09-24 14:18

from django.conf import settings
import django.contrib.postgres.fields.jsonb
from django.db import migrations, models
import django.db.models.deletion
import django_extensions.db.fields
import simple_history.models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='Action',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', django_extensions.db.fields.CreationDateTimeField(auto_now_add=True, verbose_name='created')),
                ('modified', django_extensions.db.fields.ModificationDateTimeField(auto_now=True, verbose_name='modified')),
                ('action_method', models.CharField(max_length=140)),
                ('action_kwargs', django.contrib.postgres.fields.jsonb.JSONField(default={}, help_text='kwargs to pass action method')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='BusinessRuleSet',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', django_extensions.db.fields.CreationDateTimeField(auto_now_add=True, verbose_name='created')),
                ('modified', django_extensions.db.fields.ModificationDateTimeField(auto_now=True, verbose_name='modified')),
                ('enabled', models.BooleanField(default=False, help_text='Should this rule set execute?')),
                ('name', models.CharField(help_text='A descriptive name so we know what this ruleset does', max_length=140)),
                ('notes', models.TextField(blank=True, null=True)),
                ('signal', models.CharField(help_text='import path of a signal that shoudl fire this ruleset', max_length=240)),
                ('signal_kwargs', django.contrib.postgres.fields.jsonb.JSONField(default={}, help_text="kwargs to pass to signal registration ({'sender': 'Model' would be a common pattern here})")),
                ('stop_on_first_trigger', models.BooleanField(default=False, help_text='Should this ruleset stop on the first triggered block?')),
                ('variables_class', models.CharField(help_text='Import path and class for Variables', max_length=140)),
                ('actions_class', models.CharField(help_text='Import path and class for Actions', max_length=140)),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='Condition',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', django_extensions.db.fields.CreationDateTimeField(auto_now_add=True, verbose_name='created')),
                ('modified', django_extensions.db.fields.ModificationDateTimeField(auto_now=True, verbose_name='modified')),
                ('property_method', models.CharField(max_length=140)),
                ('property_kwargs', django.contrib.postgres.fields.jsonb.JSONField(default={}, help_text='kwargs to pass to the method that gets the left hand side of our comparison')),
                ('comparison_operator', models.CharField(max_length=140)),
                ('value', models.TextField()),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='ConditionSet',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', django_extensions.db.fields.CreationDateTimeField(auto_now_add=True, verbose_name='created')),
                ('modified', django_extensions.db.fields.ModificationDateTimeField(auto_now=True, verbose_name='modified')),
                ('name', models.CharField(help_text='A descriptive name for this group of conditions', max_length=140)),
                ('combination_rule', models.CharField(choices=[('any', 'any'), ('all', 'all')], max_length=10)),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='HistoricalAction',
            fields=[
                ('id', models.IntegerField(auto_created=True, blank=True, db_index=True, verbose_name='ID')),
                ('created', django_extensions.db.fields.CreationDateTimeField(auto_now_add=True, verbose_name='created')),
                ('modified', django_extensions.db.fields.ModificationDateTimeField(auto_now=True, verbose_name='modified')),
                ('action_method', models.CharField(max_length=140)),
                ('action_kwargs', django.contrib.postgres.fields.jsonb.JSONField(default={}, help_text='kwargs to pass action method')),
                ('history_id', models.AutoField(primary_key=True, serialize=False)),
                ('history_change_reason', models.CharField(max_length=100, null=True)),
                ('history_date', models.DateTimeField()),
                ('history_type', models.CharField(choices=[('+', 'Created'), ('~', 'Changed'), ('-', 'Deleted')], max_length=1)),
            ],
            options={
                'verbose_name': 'historical action',
                'ordering': ('-history_date', '-history_id'),
                'get_latest_by': 'history_date',
            },
            bases=(simple_history.models.HistoricalChanges, models.Model),
        ),
        migrations.CreateModel(
            name='HistoricalBusinessRuleSet',
            fields=[
                ('id', models.IntegerField(auto_created=True, blank=True, db_index=True, verbose_name='ID')),
                ('created', django_extensions.db.fields.CreationDateTimeField(auto_now_add=True, verbose_name='created')),
                ('modified', django_extensions.db.fields.ModificationDateTimeField(auto_now=True, verbose_name='modified')),
                ('enabled', models.BooleanField(default=False, help_text='Should this rule set execute?')),
                ('name', models.CharField(help_text='A descriptive name so we know what this ruleset does', max_length=140)),
                ('notes', models.TextField(blank=True, null=True)),
                ('signal', models.CharField(help_text='import path of a signal that shoudl fire this ruleset', max_length=240)),
                ('signal_kwargs', django.contrib.postgres.fields.jsonb.JSONField(default={}, help_text="kwargs to pass to signal registration ({'sender': 'Model' would be a common pattern here})")),
                ('stop_on_first_trigger', models.BooleanField(default=False, help_text='Should this ruleset stop on the first triggered block?')),
                ('variables_class', models.CharField(help_text='Import path and class for Variables', max_length=140)),
                ('actions_class', models.CharField(help_text='Import path and class for Actions', max_length=140)),
                ('history_id', models.AutoField(primary_key=True, serialize=False)),
                ('history_change_reason', models.CharField(max_length=100, null=True)),
                ('history_date', models.DateTimeField()),
                ('history_type', models.CharField(choices=[('+', 'Created'), ('~', 'Changed'), ('-', 'Deleted')], max_length=1)),
                ('history_user', models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='+', to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'verbose_name': 'historical business rule set',
                'ordering': ('-history_date', '-history_id'),
                'get_latest_by': 'history_date',
            },
            bases=(simple_history.models.HistoricalChanges, models.Model),
        ),
        migrations.CreateModel(
            name='HistoricalCondition',
            fields=[
                ('id', models.IntegerField(auto_created=True, blank=True, db_index=True, verbose_name='ID')),
                ('created', django_extensions.db.fields.CreationDateTimeField(auto_now_add=True, verbose_name='created')),
                ('modified', django_extensions.db.fields.ModificationDateTimeField(auto_now=True, verbose_name='modified')),
                ('property_method', models.CharField(max_length=140)),
                ('property_kwargs', django.contrib.postgres.fields.jsonb.JSONField(default={}, help_text='kwargs to pass to the method that gets the left hand side of our comparison')),
                ('comparison_operator', models.CharField(max_length=140)),
                ('value', models.TextField()),
                ('history_id', models.AutoField(primary_key=True, serialize=False)),
                ('history_change_reason', models.CharField(max_length=100, null=True)),
                ('history_date', models.DateTimeField()),
                ('history_type', models.CharField(choices=[('+', 'Created'), ('~', 'Changed'), ('-', 'Deleted')], max_length=1)),
                ('group', models.ForeignKey(blank=True, db_constraint=False, null=True, on_delete=django.db.models.deletion.DO_NOTHING, related_name='+', to='dynamic_business_rules.ConditionSet')),
                ('history_user', models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='+', to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'verbose_name': 'historical condition',
                'ordering': ('-history_date', '-history_id'),
                'get_latest_by': 'history_date',
            },
            bases=(simple_history.models.HistoricalChanges, models.Model),
        ),
        migrations.CreateModel(
            name='HistoricalConditionSet',
            fields=[
                ('id', models.IntegerField(auto_created=True, blank=True, db_index=True, verbose_name='ID')),
                ('created', django_extensions.db.fields.CreationDateTimeField(auto_now_add=True, verbose_name='created')),
                ('modified', django_extensions.db.fields.ModificationDateTimeField(auto_now=True, verbose_name='modified')),
                ('name', models.CharField(help_text='A descriptive name for this group of conditions', max_length=140)),
                ('combination_rule', models.CharField(choices=[('any', 'any'), ('all', 'all')], max_length=10)),
                ('history_id', models.AutoField(primary_key=True, serialize=False)),
                ('history_change_reason', models.CharField(max_length=100, null=True)),
                ('history_date', models.DateTimeField()),
                ('history_type', models.CharField(choices=[('+', 'Created'), ('~', 'Changed'), ('-', 'Deleted')], max_length=1)),
            ],
            options={
                'verbose_name': 'historical condition set',
                'ordering': ('-history_date', '-history_id'),
                'get_latest_by': 'history_date',
            },
            bases=(simple_history.models.HistoricalChanges, models.Model),
        ),
        migrations.CreateModel(
            name='HistoricalRuleBlock',
            fields=[
                ('id', models.IntegerField(auto_created=True, blank=True, db_index=True, verbose_name='ID')),
                ('created', django_extensions.db.fields.CreationDateTimeField(auto_now_add=True, verbose_name='created')),
                ('modified', django_extensions.db.fields.ModificationDateTimeField(auto_now=True, verbose_name='modified')),
                ('name', models.CharField(help_text='Description of what this block does', max_length=140)),
                ('enabled', models.BooleanField(default=False, help_text='Should this block be executed as a part of the rule set?')),
                ('history_id', models.AutoField(primary_key=True, serialize=False)),
                ('history_change_reason', models.CharField(max_length=100, null=True)),
                ('history_date', models.DateTimeField()),
                ('history_type', models.CharField(choices=[('+', 'Created'), ('~', 'Changed'), ('-', 'Deleted')], max_length=1)),
                ('history_user', models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='+', to=settings.AUTH_USER_MODEL)),
                ('rule_set', models.ForeignKey(blank=True, db_constraint=False, null=True, on_delete=django.db.models.deletion.DO_NOTHING, related_name='+', to='dynamic_business_rules.BusinessRuleSet')),
            ],
            options={
                'verbose_name': 'historical rule block',
                'ordering': ('-history_date', '-history_id'),
                'get_latest_by': 'history_date',
            },
            bases=(simple_history.models.HistoricalChanges, models.Model),
        ),
        migrations.CreateModel(
            name='RuleBlock',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', django_extensions.db.fields.CreationDateTimeField(auto_now_add=True, verbose_name='created')),
                ('modified', django_extensions.db.fields.ModificationDateTimeField(auto_now=True, verbose_name='modified')),
                ('name', models.CharField(help_text='Description of what this block does', max_length=140)),
                ('enabled', models.BooleanField(default=False, help_text='Should this block be executed as a part of the rule set?')),
                ('rule_set', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='dynamic_business_rules.BusinessRuleSet')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.AddField(
            model_name='historicalconditionset',
            name='block',
            field=models.ForeignKey(blank=True, db_constraint=False, null=True, on_delete=django.db.models.deletion.DO_NOTHING, related_name='+', to='dynamic_business_rules.RuleBlock'),
        ),
        migrations.AddField(
            model_name='historicalconditionset',
            name='history_user',
            field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='+', to=settings.AUTH_USER_MODEL),
        ),
        migrations.AddField(
            model_name='historicalconditionset',
            name='parent_condition',
            field=models.ForeignKey(blank=True, db_constraint=False, null=True, on_delete=django.db.models.deletion.DO_NOTHING, related_name='+', to='dynamic_business_rules.ConditionSet'),
        ),
        migrations.AddField(
            model_name='historicalaction',
            name='block',
            field=models.ForeignKey(blank=True, db_constraint=False, null=True, on_delete=django.db.models.deletion.DO_NOTHING, related_name='+', to='dynamic_business_rules.RuleBlock'),
        ),
        migrations.AddField(
            model_name='historicalaction',
            name='history_user',
            field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='+', to=settings.AUTH_USER_MODEL),
        ),
        migrations.AddField(
            model_name='conditionset',
            name='block',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='condition_sets', to='dynamic_business_rules.RuleBlock'),
        ),
        migrations.AddField(
            model_name='conditionset',
            name='parent_condition',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='child_condition_sets', to='dynamic_business_rules.ConditionSet'),
        ),
        migrations.AddField(
            model_name='condition',
            name='group',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='comditions', to='dynamic_business_rules.ConditionSet'),
        ),
        migrations.AddField(
            model_name='action',
            name='block',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='actions', to='dynamic_business_rules.RuleBlock'),
        ),
    ]



from django.contrib import admin

import nested_admin

from .models import BusinessRuleSet, RuleBlock, ConditionSet, Condition, Action
from .forms import BusinessRuleSetAdminForm, ConditionAdminForm, ActionAdminForm


class ConditionInline(nested_admin.NestedStackedInline):
    model = Condition
    extra = 0
    form = ConditionAdminForm
    fieldsets = [
        (None, {
            "fields": [
                ('property_method', 'comparison_operator', 'value'),
                'property_kwargs'
            ]
        })
    ]


class ConditionSetInline(nested_admin.NestedStackedInline):
    model = ConditionSet
    inlines = [ConditionInline]
    extra = 0
    fieldsets = [
        (None, {
            "fields": [('name', 'combination_rule', 'parent_condition')]
        })
    ]


class ActionInline(nested_admin.NestedStackedInline):
    model = Action
    extra = 0
    form = ActionAdminForm


class RuleBlockInline(nested_admin.NestedStackedInline):
    model = RuleBlock
    inlines = [ConditionSetInline, ActionInline]
    extra = 0
    fieldsets = [
        (None, {
            "fields": [('name', 'enabled')]
        })
    ]


@admin.register(BusinessRuleSet)
class BusinessRuleSetAdmin(nested_admin.NestedModelAdmin):
    inlines = [RuleBlockInline]
    form = BusinessRuleSetAdminForm
    fieldsets = [
        (None, {
            "fields": [
                ('name', 'enabled'),
                'notes',
                ('signal', 'signal_kwargs'),
                ('variables_class', 'actions_class'),
                'stop_on_first_trigger',
            ],
        })
    ]
    list_display = [
        'name',
        'signal',
        'signal_kwargs',
        'variables_class',
        'actions_class'
    ]



# Generated by Django 2.0.7 on 2018-09-30 01:17

import django.contrib.postgres.fields.jsonb
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('dynamic_business_rules', '0001_initial'),
    ]

    operations = [
        migrations.AlterField(
            model_name='businessruleset',
            name='signal_kwargs',
            field=django.contrib.postgres.fields.jsonb.JSONField(blank=True, default={}, help_text="kwargs to pass to signal registration ({'sender': 'Model' would be a common pattern here})"),
        ),
        migrations.AlterField(
            model_name='condition',
            name='property_kwargs',
            field=django.contrib.postgres.fields.jsonb.JSONField(blank=True, default={}, help_text='kwargs to pass to the method that gets the left hand side of our comparison'),
        ),
        migrations.AlterField(
            model_name='condition',
            name='property_method',
            field=models.CharField(blank=True, max_length=140, null=True),
        ),
        migrations.AlterField(
            model_name='condition',
            name='value',
            field=models.TextField(blank=True, null=True),
        ),
        migrations.AlterField(
            model_name='conditionset',
            name='combination_rule',
            field=models.CharField(choices=[('any', 'any'), ('all', 'all')], default='all', max_length=10),
        ),
        migrations.AlterField(
            model_name='historicalbusinessruleset',
            name='signal_kwargs',
            field=django.contrib.postgres.fields.jsonb.JSONField(blank=True, default={}, help_text="kwargs to pass to signal registration ({'sender': 'Model' would be a common pattern here})"),
        ),
        migrations.AlterField(
            model_name='historicalcondition',
            name='property_kwargs',
            field=django.contrib.postgres.fields.jsonb.JSONField(blank=True, default={}, help_text='kwargs to pass to the method that gets the left hand side of our comparison'),
        ),
        migrations.AlterField(
            model_name='historicalcondition',
            name='property_method',
            field=models.CharField(blank=True, max_length=140, null=True),
        ),
        migrations.AlterField(
            model_name='historicalcondition',
            name='value',
            field=models.TextField(blank=True, null=True),
        ),
        migrations.AlterField(
            model_name='historicalconditionset',
            name='combination_rule',
            field=models.CharField(choices=[('any', 'any'), ('all', 'all')], default='all', max_length=10),
        ),
    ]



from django_extensions.db.models import TimeStampedModel
from simple_history.models import HistoricalRecords

from django.db import models
from django.utils.module_loading import import_string
from django.contrib.postgres.fields import JSONField


class BaseModel(TimeStampedModel):
    """
    By default all models should have timestamp info and history
    """
    class Meta:
        abstract = True

    history = HistoricalRecords(inherit=True)


class BusinessRuleSet(BaseModel):
    """
    A collection of rules that are evaluated in order, if
    stop_on_first_trigger is true the rule set will short circuit as
    soon as the first rule set evaluates to true, otherwise all rules
    in the rule set will be evaluated.
    """
    enabled = models.BooleanField(
        default=False,
        help_text="Should this rule set execute?"
    )
    name = models.CharField(
        max_length=140,
        help_text="A descriptive name so we know what this ruleset does"
    )
    notes = models.TextField(
        null=True,
        blank=True
    )
    signal = models.CharField(
        max_length=240,
        help_text="import path of a signal that shoudl fire this ruleset",
    )
    signal_kwargs = JSONField(
        default={},
        blank=True,
        help_text="kwargs to pass to signal registration ({'sender': 'Model' would be a common pattern here})"
    )
    stop_on_first_trigger = models.BooleanField(
        default=False,
        help_text="Should this ruleset stop on the first triggered block?"
    )
    variables_class = models.CharField(
        max_length=140,
        help_text="Import path and class for Variables"
    )
    actions_class = models.CharField(
        max_length=140,
        help_text="Import path and class for Actions"
    )

    def __str__(self):
        return self.name

    @property
    def variable_class_object(self):
        return import_string(self.variables_class)

    @property
    def action_class_object(self):
        return import_string(self.actions_class)

    def signal_callback(self, *args, **kwargs):
        if not self.enabled:
            return

        # NOTE: Variables and Action classes need to accept args and kwargs
        # from signal callback as arguments to their constructor
        variables_instance = self.variable_class_object(*args, **kwargs)
        actions_instance = self.action_class_object(*args, **kwargs)

        for rule_block in self.rule_blocks.filter(enabled=True):
            result = rule_block.run(variables_instance, actions_instance)
            if result is True and self.stop_on_first_trigger:
                return


class RuleBlock(BaseModel):
    rule_set = models.ForeignKey(
        BusinessRuleSet,
        on_delete=models.CASCADE
    )
    name = models.CharField(
        max_length=140,
        help_text="Description of what this block does"
    )
    enabled = models.BooleanField(
        default=False,
        help_text="Should this block be executed as a part of the rule set?"
    )

    def __str__(self):
        return self.name

    def run(self, variables_instance, actions_instance):
        if self.all_conditions_true(variables_instance, actions_instance):
            self.run_actions(variables_instance, actions_instance)
            return True
        else:
            return False

    def all_conditions_true(self, variables_instance):
        return all(map(lambda group: group.evaluate(variables_instance), self.condition_sets.all()))

    def run_actions(self, variables_instance, actions_instance):
        for action in self.actions.all():
            action.run(variables_instance, actions_instance)


class ConditionSet(BaseModel):
    block = models.ForeignKey(
        RuleBlock,
        on_delete=models.CASCADE,
        related_name='condition_sets'
    )
    name = models.CharField(
        help_text="A descriptive name for this group of conditions",
        max_length=140
    )
    parent_condition = models.ForeignKey(
        'self',
        null=True,
        blank=True,
        related_name='child_condition_sets',
        on_delete=models.CASCADE
    )
    combination_rule = models.CharField(
        max_length=10,
        choices=[
            ('any', 'any'),
            ('all', 'all')
        ],
        default='all'
    )

    def __str__(self):
        return self.name

    def evaluate(self, variables_instance):
        if self.combination_rule == 'any':
            return any(self.depth_first_evaluate_results(variables_instance))
        if self.combination_rule == 'all':
            return any(self.depth_first_evaluate_results(variables_instance))

        raise ValueError('could not evaluate ConditionSet with combination rule %s', self.combination_rule)

    def depth_first_evaluate_results(self, variables_instance):
        return self.child_set_results(variables_instance) + self.child_condition_results(variables_instance)

    def child_set_results(self, variables_instance):
        return list(
            map(
                lambda condition_set: condition_set.evaluate(variables_instance),
                self.child_condition_sets.all()
            )
        )

    def child_condition_results(self, variables_instance):
        return list(map(lambda condition: condition.evaluate(variables_instance), self.conditions.all()))


class Condition(BaseModel):
    group = models.ForeignKey(
        ConditionSet,
        on_delete=models.CASCADE,
        related_name='comditions'
    )
    property_method = models.CharField(
        max_length=140,
        null=True,
        blank=True
    )
    property_kwargs = JSONField(
        default={},
        help_text="kwargs to pass to the method that gets the left hand side of our comparison",
        blank=True
    )
    comparison_operator = models.CharField(
        max_length=140
    )
    value = models.TextField(
        blank=True,
        null=True
    )

    def __str__(self):
        if self.property_method is None:
            return '-'

        return self.property_method

    @property
    def operator_instance(self):
        return import_string(self.comparison_operator)

    @property
    def variable_class_object(self):
        return self.group.block.rule_set.variable_class_object

    @property
    def property_method_instance(self):
        if self.property_method and hasattr(self.variable_class_object, self.property_method):
            return getattr(self.variable_class_object, self.property_method)

    @property
    def comparison_operator_choices(self):
        if self.property_method_instance and hasattr(self.property_method_instance, 'operators'):
            return self.property_method_instance.operators
        else:
            return self.get_default_comparison_operators()

    @staticmethod
    def get_default_comparison_operators():
        return [
            'operator.lt',
            'operator.le',
            'operator.eq',
            'operator.ne',
            'operator.ge',
            'operator.gt',
            'operator.not_',
            'operator.truth',
            'operator.is',
            'operator.is_not',
            'str.startswith',
            'str.endswith'
        ]

    @property
    def variable_properties(self):
        return [
            method
            for method
            in dir(self.variable_class_object)
            if not method.startswith('__')
        ]

    def evaluate(self, variable_instance):
        return self.operator_instance(
            getattr(variable_instance, self.property_method)(**self.property_kwargs),
            self.value
        )


class Action(BaseModel):
    block = models.ForeignKey(
        RuleBlock,
        on_delete=models.CASCADE,
        related_name='actions'
    )
    action_method = models.CharField(
        max_length=140
    )
    action_kwargs = JSONField(
        default={},
        help_text="kwargs to pass action method"
    )

    def __str__(self):
        return self.action_method

    def run(self, variables_instance, actions_instance):
        # All action methods need to take variables as their first argument
        getattr(actions_instance, self.action_method)(variables_instance, self.action_kwargs)

    @property
    def action_class_object(self):
        return self.block.rule_set.action_class_object

    @property
    def action_method_choices(self):
        return [
            method
            for method
            in dir(self.action_class_object)
            if not method.startswith('__')
        ]



import inspect

from django import forms
from django.conf import settings
from django.utils.module_loading import import_string

from .models import BusinessRuleSet, Condition, Action


def module_exists(module_name):
    try:
        __import__(module_name)
    except ImportError:
        return False
    else:
        return True


def modules_with_variables():
    return [
        app
        for app
        in map(lambda app: f"{app}.dynamic_rule_variables", settings.INSTALLED_APPS)
        if module_exists(app)
    ]


def modules_with_actions():
    return [
        app
        for app
        in map(lambda app: f"{app}.dynamic_rule_actions", settings.INSTALLED_APPS)
        if module_exists(app)
    ]


def classes_from_module_list(module_list):
    for module in module_list:
        for name, obj in inspect.getmembers(import_string(module)):
            if inspect.isclass(obj):
                class_name = obj.__name__
                yield f"{module}.{class_name}"


class BusinessRuleSetAdminForm(forms.ModelForm):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.fields['variables_class'] = forms.ChoiceField(
            choices=self.variables_class_choices,
            widget=forms.Select(attrs={'style': 'font-size:10px'})
        )
        self.fields['actions_class'] = forms.ChoiceField(
            choices=self.actions_class_choices,
            widget=forms.Select(attrs={'style': 'font-size:10px'})
        )
        self.fields['signal_kwargs'] = forms.CharField(required=False)

    @property
    def variables_class_choices(self):
        variable_classes = list(classes_from_module_list(modules_with_variables()))

        return zip(variable_classes, variable_classes)

    @property
    def actions_class_choices(self):
        action_classes = list(classes_from_module_list(modules_with_actions()))

        return zip(action_classes, action_classes)

    class Meta:
        model = BusinessRuleSet
        fields = [
            'name',
            'enabled',
            'notes',
            'signal',
            'signal_kwargs',
            'variables_class',
            'actions_class',
            'stop_on_first_trigger'
        ]


class ConditionAdminForm(forms.ModelForm):
    class Meta:
        model = Condition
        fields = [
            'group',
            'property_method',
            'property_kwargs',
            'comparison_operator',
            'value'
        ]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        comparison_choices = ['op.eq', 'op.ne']

        if kwargs.get('instance') is not None:
            instance = kwargs['instance']
            properties = instance.variable_properties

            self.fields['property_method'] = forms.ChoiceField(
                choices=zip(properties, properties)
            )

            self.fields['property_kwargs'] = forms.CharField()
            self.fields['value'] = forms.CharField(required=False)

            comparison_choices = instance.comparison_operator_choices

        self.fields['comparison_operator'] = forms.ChoiceField(
            choices=zip(comparison_choices, comparison_choices)
        )


class ActionAdminForm(forms.ModelForm):
    class Meta:
        model = Action
        fields = [
            'block',
            'action_method',
            'action_kwargs'
        ]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.fields['action_kwargs'] = forms.CharField(required=False)

        if kwargs.get('instance') is not None:
            instance = kwargs['instance']

            self.fields['action_method'] = forms.ChoiceField(
                choices=zip(instance.action_method_choices, instance.action_method_choices)
            )
